#pragma once

#include "mindsensorsDriver.h"
#include "can_sd540.h"

#include "mindsensorsConfidentialDriver.h"

#include "HAL/Types.h"
#include "HAL/handles/IndexedHandleResource.h"

#include <string>
#include <chrono> /* for GetBatteryVoltage grace period */
#include <atomic>

#define CANSD540_Handle HAL_Handle

namespace mindsensors {

class CANSD540Driver : protected mindsensorsDriver, mindsensorsConfidentialDriver {
public:
    static std::string GetLibraryVersion();

    // enums copied in user class CANSD540.h to be used with CANSD540::Coast instead of CANSD540Driver::Coast
    enum StopMode : uint8_t {
        Coast = 0,
        Brake = 1
    };

    enum LimitSwitchBehavior : uint8_t {
        NormallyOpen = 0,
        NormallyClosed = 1
    };

    enum State : uint8_t {
        Enabled = 0,
        NotFound = 1,
        OldFirmware = 2
    };

    CANSD540Driver(int8_t deviceNumber, int32_t* status);

        uint8_t GetDeviceID(int32_t* status) const;
    std::string GetDeviceName(int32_t* status) const;
    std::string GetFirmwareVersion(int32_t* status) const;
    std::string GetHardwareVersion(int32_t* status) const;
    std::string GetBootloaderVersion(int32_t* status) const;
    std::string GetSerialNumber(int32_t* status) const;

    void BlinkLED(uint8_t seconds, int32_t* status);

    void Set(double speed, int32_t* status);
    double Get(int32_t* status) const;

    void SetInverted(HAL_Bool isInverted, int32_t* status);
    HAL_Bool GetInverted(int32_t* status) const;

    void SetStopMode(StopMode stopMode, int32_t* status);
    StopMode GetStopMode(int32_t* status) const;

    void EnableLimitSwitches(HAL_Bool limitSwitchesEnabled, int32_t* status);
    HAL_Bool GetLimitSwitchesEnabled(int32_t* status) const;

    void SetForwardLimitSwitch(LimitSwitchBehavior behavior, int32_t* status);
    LimitSwitchBehavior GetForwardLimitSwitch(int32_t* status) const;

    void SetReverseLimitSwitch(LimitSwitchBehavior behavior, int32_t* status);
    LimitSwitchBehavior GetReverseLimitSwitch(int32_t* status) const;

    void SetVoltageRamp(uint8_t voltageRamp, int32_t* status);
    uint8_t GetVoltageRamp(int32_t* status) const;

    double GetBatteryVoltage(int32_t* status);

    HAL_Bool GetForwardLimitSwitchState(int32_t* status);
    HAL_Bool GetReverseLimitSwitchState(int32_t* status);

protected:
        uint8_t m_deviceID;
    std::string m_deviceName;
    std::string m_firmwareVersion;
    std::string m_hardwareVersion;
    std::string m_bootloaderVersion;
    std::string m_serialNumber;

    uint8_t lastStatusPacketData[8];
    std::chrono::time_point<std::chrono::system_clock> lastStatusPacketTime;

private:
    static std::atomic<hal::IndexedHandleResource<CANSD540_Handle, uint8_t, 63, hal::HAL_HandleEnum::Vendor>*> cansd540Handles;
    CANSD540_Handle m_resourceHandle;

    State state = State::Enabled;
    void DisabledWarning(std::string methodName) const;
};

} // namespace mindsensors

extern "C" {

const char* CANSD540_GetLibraryVersion();

int CANSD540_Constructor(int8_t deviceNumber, int32_t* status);
void CANSD540_Destructor(CANSD540_Handle handle);

    uint8_t CANSD540_GetDeviceID(CANSD540_Handle handle, int32_t* status);
const char* CANSD540_GetDeviceName(CANSD540_Handle handle, int32_t* status);
const char* CANSD540_GetFirmwareVersion(CANSD540_Handle handle, int32_t* status);
const char* CANSD540_GetHardwareVersion(CANSD540_Handle handle, int32_t* status);
const char* CANSD540_GetBootloaderVersion(CANSD540_Handle handle, int32_t* status);
const char* CANSD540_GetSerialNumber(CANSD540_Handle handle, int32_t* status);

void CANSD540_BlinkLED(CANSD540_Handle handle, uint8_t seconds, int32_t* status);

void CANSD540_Set(CANSD540_Handle handle, double speed, int32_t* status);
double CANSD540_Get(CANSD540_Handle handle, int32_t* status);

void CANSD540_SetInverted(CANSD540_Handle handle, HAL_Bool isInverted, int32_t* status);
HAL_Bool CANSD540_GetInverted(CANSD540_Handle handle, int32_t* status);

void CANSD540_SetStopMode(CANSD540_Handle handle, mindsensors::CANSD540Driver::StopMode stopMode, int32_t* status);
mindsensors::CANSD540Driver::StopMode CANSD540_GetStopMode(CANSD540_Handle handle, int32_t* status);

void CANSD540_EnableLimitSwitches(CANSD540_Handle handle, HAL_Bool limitSwitchesEnabled, int32_t* status);
HAL_Bool CANSD540_GetLimitSwitchesEnabled(CANSD540_Handle handle, int32_t* status);

void CANSD540_SetForwardLimitSwitch(CANSD540_Handle handle, mindsensors::CANSD540Driver::LimitSwitchBehavior behavior, int32_t* status);
mindsensors::CANSD540Driver::LimitSwitchBehavior CANSD540_GetForwardLimitSwitch(CANSD540_Handle handle, int32_t* status);

void CANSD540_SetReverseLimitSwitch(CANSD540_Handle handle, mindsensors::CANSD540Driver::LimitSwitchBehavior behavior, int32_t* status);
mindsensors::CANSD540Driver::LimitSwitchBehavior CANSD540_GetReverseLimitSwitch(CANSD540_Handle handle, int32_t* status);

void CANSD540_SetVoltageRamp(CANSD540_Handle handle, uint8_t voltageRamp, int32_t* status);
uint8_t CANSD540_GetVoltageRamp(CANSD540_Handle handle, int32_t* status);

double CANSD540_GetBatteryVoltage(CANSD540_Handle handle, int32_t* status);

HAL_Bool CANSD540_GetForwardLimitSwitchState(CANSD540_Handle handle, int32_t* status);
HAL_Bool CANSD540_GetReverseLimitSwitchState(CANSD540_Handle handle, int32_t* status);

} // extern "C"
